part of orm_client;

class EntityManager {
  
  static EntityManager _instance;
  
  static String host = '127.0.0.1';
  static String port = '8080';
  
  List<EntityScan> _entityScans = new List<EntityScan>();
  Map<String, Map<String, Entity>> _spawnRegistry = new Map<String, Map<String, Entity>>();
  
  //---------------------------------
  //
  // Singleton Constructor
  //
  //---------------------------------
  
  EntityManager._construct();

  factory EntityManager() {
    if (_instance == null) {
      _instance = new EntityManager._construct();
    }

    return _instance;
  }
  
  Entity _spawn(Map rawData) {
    final String type = rawData['__type__'];
    final Symbol typeSymbol = new Symbol(type);
    Entity entity;
    
    Iterable<EntityScan> existingScans = _entityScans.where(
        (EntityScan scan) => (scan.qualifiedName.toString() == typeSymbol.toString())
    );
    
    if (existingScans.length > 0) {
      EntityScan scan = existingScans.first;
      
      scan.classMirror.constructors.forEach(
        (Symbol K, MethodMirror v) {
          entity = scan.classMirror.newInstance(v.constructorName, []).reflectee;
        }
      );
      
      if (entity != null) {
        entity._scan._proxies.forEach(
          (_ProxyEntry entry) {
            dynamic entryValue = rawData[entry.property];
            
            if (entryValue is Map) {
              entry.proxy._value = _spawn(entryValue);
            } else {
              entry.proxy._value = entryValue;
            }
          }
        );
      }
      
      Iterable<_ProxyEntry> idProxies = entity._scan._proxies.where(
        (_ProxyEntry entry) => entry.proxy.isId   
      );
      
      // build ID map
      List<String> idList = <String>[];
      
      idProxies.forEach(
        (_ProxyEntry entry) => idList.add(_keyBuilder(entry))
      );
      
      String key = idList.join('??');
      
      return _existingFromSpawnRegistry(type, key, entity);
    }
    
    throw new ArgumentError('Scan for entity not found');
    
    return null;
  }
  
  Entity _existingFromSpawnRegistry(String type, String key, Entity candidateEntity) {
    Map<String, Entity> typeRegistry;
    
    if (!_spawnRegistry.containsKey(type)) {
      _spawnRegistry[type] = new Map<String, Entity>();
    }
    
    typeRegistry = _spawnRegistry[type];
    
    if (typeRegistry.containsKey(key)) {
      return typeRegistry[key];
    }
    
    typeRegistry[key] = candidateEntity;
    
    return candidateEntity;
  }
  
  String _keyBuilder(_ProxyEntry entry) {
    String key;
    
    if (entry.proxy.value is Entity) {
      List<String> idList = <String>[];
      Entity entity = entry.proxy.value;
      
      Iterable<_ProxyEntry> idProxies = entity._scan._proxies.where(
          (_ProxyEntry entry) => entry.proxy.isId   
      );
      
      idProxies.forEach(
        (_ProxyEntry entry) => idList.add(
            '${entry.property}?${entry.proxy.value}'    
        )
      );
      
      key = idList.join('????');
    } else {
      key = '${entry.property}?${entry.proxy.value}';
    }
    
    return key;
  }
  
  List<Type> _reflectedTypes = new List<Type>();
  
  EntityScan _getExistingScan(Type forType) {
    Iterable<EntityScan> existingScans = _entityScans.where(
      (EntityScan scan) => (scan.entityType == forType)
    );
    
    if (existingScans.length > 0) {
      return existingScans.first;
    }
    
    return null;
  }
  
  EntityScan scan(Type forType) {
    EntityScan scan = _getExistingScan(forType);
    
    if(scan != null) {
      return scan;
    }
    
    scan = new EntityScan();
    
    Property property;
    ClassMirror classMirror = reflectClass(forType);
    List<InstanceMirror> metadata = <InstanceMirror>[];
    Map<Symbol, Mirror> members = new Map<Symbol, Mirror>();
    
    scan.classMirror = classMirror;
    
    classMirror.metadata.forEach(
        (InstanceMirror metadata) {
          if (metadata.reflectee is Ref) {
            scan.qualifiedName = new Symbol((metadata.reflectee as Ref).path);
          }
        }
    );
    
    metadata.addAll(classMirror.metadata);
    members.addAll(classMirror.members);
    
    classMirror = classMirror.superclass;
    
    while (classMirror.qualifiedName.toString() != 'Symbol("orm_client.Entity")') {
      metadata.addAll(classMirror.metadata);
      members.addAll(classMirror.members);
      
      classMirror = classMirror.superclass;
    }
    
    metadata.forEach(
      (InstanceMirror metadata) {
        if (metadata.reflectee is Immutable) {
          scan.isMutableEntity = false;
        }
      }
    );
    
    members.forEach(
      (Symbol K, Mirror v) {
        if (v is VariableMirror) {
          v.metadata.forEach(
              (InstanceMirror metadata) {
                if (metadata.reflectee is Property) {
                  property = metadata.reflectee as Property;
                  
                  scan.addProxy(property.property, K, v.metadata);
                }
              }
          );
        }
      }
    );
    
    _entityScans.add(scan);
    
    return scan;
  }
  
  void _initialize(Entity entity) {
    InstanceMirror instanceMirror = reflect(entity);
    Proxy proxy;
    
    entity._scan._proxies.forEach(
      (_ProxyEntry entry) {
        proxy = new Proxy._construct(null, true);
        
        proxy.property = entry.property;
        
        entry.metadata.forEach(
            (InstanceMirror metadata) {
              if (metadata.reflectee is Id) {
                proxy.isId = true;
              } else if (metadata.reflectee is Transient) {
                proxy.isTransient = true;
              } else if (metadata.reflectee is DefaultValue) {
                proxy._value = (metadata.reflectee as DefaultValue).value;
              } else if (
                  !entity._scan.isMutableEntity ||
                  (metadata.reflectee is Immutable)
                ) {
                proxy.isMutable = false;
              }
            }
        );
        
        entry.proxy = proxy;
        
        instanceMirror.setField(entry.symbol, proxy);
      }
    );
  }
  
  Future<Type> test_handler() {
    Completer completer = new Completer();
    
    Timer.run(
      () => completer.complete('ALL_OK') 
    );
    
    return completer.future;
  }
  
}