library orm_client;

import 'dart:async';
import 'dart:mirrors';
import 'dart:core';
import 'dart:html';
import 'dart:json';

part 'core/entity_manager.dart';
part 'core/entity_scan.dart';

part 'domain/entity.dart';
part 'domain/meta.dart';

part 'types/types.dart';

Future<List<Entity>> ormEntityLoad(String entityName, {Map<String, dynamic> where}) {
  Completer completer = new Completer();
  Future<HttpRequest> response;
  
  if (where != null) {
    List<String> whereClauses = <String>[];
    
    where.forEach(
      (String whereName, dynamic whereValue) => whereClauses.add('${whereName}=${whereValue}')    
    );
    
    response = HttpRequest.request('http://${EntityManager.host}:${EntityManager.port}?entityName=${entityName}&${whereClauses.join('&')}');
  } else {
    response = HttpRequest.request('http://${EntityManager.host}:${EntityManager.port}?entityName=${entityName}');
  }
  
  response.then(
    (HttpRequest request) {
      List<Entity> resultList = <Entity>[];
      List<Map<String, dynamic>> result = parse(request.responseText);
      
      EntityManager entityManager = new EntityManager();
      
      result.forEach(
          (Map<String, dynamic> resultEntry) => resultList.add(entityManager._spawn(resultEntry))  
      );
      
      completer.complete(resultList);
    }, onError: (_) {
      print('Oops, something went wrong, are you sure that the server has started up correctly?');
    }
  );
  
  return completer.future;
}