part of orm_client;

class Proxy<T> {

  T _defaultValue;
  T _value;
  
  T get value => _value;
  set value(T newValue) {
    if (!isMutable) {
      throw new ArgumentError('$property is immutable');
    }
    
    _value = newValue;
  }
  
  String property;
  bool isId = false;
  bool isTransient = false;
  bool isMutable = true;
  
  T operator ~() => value;
  
  Proxy._construct(T v, bool m) {
    _defaultValue = v;
    
    value = v;
    
    isMutable = m;
  }
  
}

class LazyReference<T> {
  
  Function _handler;
  bool isImmediateAvailable = false;
  T _internalValue;
  
  LazyReference._construct(Function handler) {
    _handler = handler;
  }
  
  T get immediate {
    if (!isImmediateAvailable) {
      throw new ArgumentError('Reference is not loaded, use future instead');
    }
    
    return _internalValue;
  }
  
  Future<T> get future {
    final Completer completer = new Completer();
    
    if (isImmediateAvailable) {
      Timer.run(
        () => completer.complete(_internalValue) 
      );
    } else {
      // obtain value and complete
      _handler().then(
          (T value) {
            _internalValue = value;
            
            isImmediateAvailable = true;
            
            completer.complete(_internalValue);
          }
      );
    }
    
    return completer.future;
  }
  
}