import 'dart:async';
import 'dart:io';
import 'dart:json';

class JsonDatabase {
  
  Future<Map<String, dynamic>> ormEntityLoadByPK(String entityName, int id) {
    
  }
  
  Future<List<Map<String, dynamic>>> ormEntityLoad(String entityName) {
    Completer completer = new Completer();
    List<Definition> definitions = <Definition>[];
    Directory dir = new Directory('entities');
    Definition definition;
    
    Future<List<FileSystemEntity>> listing = getDefinitionFiles(dir);
    
    listing.then(
        (List<FileSystemEntity> result) {
          result.forEach(
              (File file) {
                definition = new Definition();
                
                List<String> tmp = file.path.replaceAll('\\', '.').split('.');
                
                tmp.removeLast();
                
                definition.type = tmp.join('.');
                definition.definition = parse(
                  file.readAsStringSync(encoding: Encoding.UTF_8)  
                );
                
                definitions.add(definition);
              }
          );
          
          completer.complete(
              getEntityMapList(entityName, definitions)
          );
        }
    );
    
    return completer.future;
  }
  
  List<Map<String, dynamic>> getEntityMapList(String entityName, List<Definition> definitions) {
    List<Map<String, dynamic>> list = <Map<String, dynamic>>[];
    Map<String, dynamic> tmpResult;
    Definition result = definitions.where(
        (Definition definition) => (definition.definition['name'] == entityName)    
    ).first;
    
    tmpResult = result.definition;
    
    while (tmpResult.containsKey('extends')) {
      tmpResult = definitions.where(
          (Definition definition) => (definition.definition['name'] == tmpResult['extends'])    
      ).first.definition;
      
      result.definition['properties'].addAll(tmpResult['properties']);
    }
    
    List<Map<String, dynamic>> tableRows = getTableRows(result.definition['table']);
    
    tableRows.forEach(
        (Map<String, dynamic> row) {
          Map<String, dynamic> entityMap = new Map<String, dynamic>();
          
          entityMap['__type__'] = result.type;
          
          result.definition['properties'].forEach(
              (Map<String, dynamic> propery) {
                Iterable<Definition> typeMatches = definitions.where(
                    (Definition def) => (def.definition['name'] == propery['type'])                
                );
                
                String lookupName = propery['name'];
                String lookupColumn = propery['column'];
                
                if (typeMatches.length > 0) {
                  String lookupType = propery['type'];
                  Map<String, dynamic> typeEntityMatch = typeMatches.first.definition;
                  int lookupId = row[lookupColumn];
                  
                  List<Map<String, dynamic>> result = getEntityMapList(lookupType, definitions);
                  
                  Iterable<Map<String, dynamic>> rowResults = result.where(
                      (Map<String, dynamic> resultEntry) => (resultEntry['id'] == lookupId)   
                  );
                  
                  if (rowResults.length > 0) {
                    entityMap[lookupName] = rowResults.first;
                  } else {
                    entityMap[lookupName] = null;
                  }
                } else {
                  entityMap[lookupName] = row[lookupColumn];
                }
              }
          );
          
          list.add(entityMap);
        }
    );
    
    return list;
  }
  
  List<Map<String, dynamic>> getTableRows(String tableName) {
    Completer completer = new Completer();
    File tableFile = new File('dbo/dbo_${tableName}.json');
    
    return parse(tableFile.readAsStringSync(encoding: Encoding.UTF_8))['rows'];
  }
  
  Future<List<FileSystemEntity>> getDefinitionFiles(Directory dir) {
    List<FileSystemEntity> files = <FileSystemEntity>[];
    Completer completer = new Completer();
    Stream<FileSystemEntity> lister = dir.list(recursive: false);
    
    lister.listen ( 
        (FileSystemEntity file) => 
            files.add(file),
            onDone: () => completer.complete(files)
    );
    
    return completer.future;
  }
}

class Definition {
  
  String type;
  Map<String, dynamic> definition;
  
}